#!/usr/bin/env python

# AUTOMATIC PICKER
#
# dependencies: mseed_router.py, stream_utils.py
#
# (c) Ludek Vecsey, IG CAS, 2017
#
# NOTE: beware of user unfriendly code (with mostly no comments)!
__author__ = "Ludek Vecsey"
__version__ = "v0.1beta"


import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
import os, glob
from obspy import UTCDateTime, read_inventory, Stream, Trace
from obspy.xseed import Parser
from obspy.core.event import readEvents
from obspy.core import read
from obspy.signal.trigger import plotTrigger, recSTALTA, triggerOnset
from obspy.signal.filter import envelope
from stream_utils import *
from mseed_router import read_mseed_route, stations_in_network
from scipy.optimize import curve_fit
from copy import deepcopy
import numpy as np

#--- INPUTS -------------------------------------------------
catalog = 'events.xml'

fold_home = '.'
fold_dless = os.path.join(fold_home, 'DLESS')
fold_mseed = os.path.join(fold_home, 'MSEED')
routefile = 'data_archives.txt'

version = __version__

fold_out = os.path.join(fold_home, 'PICKS_ALL', version)
fold_out_pics = os.path.join(fold_home, 'OBR', version)

if not os.path.isdir(fold_out):
    os.makedirs(fold_out)
if not os.path.isdir(fold_out_pics):
    os.makedirs(fold_out_pics)

io_figure = True          # save figures?
io_only_archive = False   # use already saved mseeds only?

phase = 'P'
mseed_offset = [-60, 60]  # read mseed around P-arrival time

# prefered sample_rate (in Hz)
sample_rate = 100.

# trigger
sta = 1    # length in sec
lta = 10   # length in sec
trig_t1 = 4.
trig_t2 = 3.

# how much peaks
peak_min = 2
peak_max = 5

# max time difference from a theoretical phase
max_theo_dif = 5.    # in sec

# station group
GRP = ['KW1', 'MANZ', 'MGBB', 'MKON', 'MROB', 'MSBB']


#--- CLASSES -------------------------------------------------------

class Correl_stats(object):
    """
    Parameters from signal correlations
    """
    def __init__(self):
        self.cor_max = []
        self.roots = []
        self.maxmins = []
        self.wave_start = []
        self.trace = Trace()
        self.picks = []
        self.noise_level = []
    def copy(self):
        return deepcopy(self)
    
    def median(self, st, icorstat=0, irange=None):
        """median of signals"""
        st1 = st.copy()
        if irange is None:
            irange = range(len(st1[0].corstats[icorstat].trace.data))  # take it all for normalization
        ##### local arrays separately
        _st = st1.select(network='GR', station='GE??')
        if len(_st) > 0:
            _signals = _st[0].corstats[icorstat].trace.data / np.linalg.norm(
                                            _st[0].corstats[icorstat].trace.data[irange])
            for i in range(1, len(_st)):
                _signals = np.vstack((_signals, _st[i].corstats[icorstat].trace.data 
                                    / np.linalg.norm(_st[i].corstats[icorstat].trace.data[irange])))
            _median = np.median(_signals, axis=0)
            st1 = Stream(traces=[tr for tr in st1 if tr not in _st])
            _st[0].corstats[icorstat].trace.data = _median
            st1.append(_st[0])
        #####
        self.trace = st1[0].corstats[icorstat].trace.copy()
        signals = st1[0].corstats[icorstat].trace.data / np.linalg.norm(st1[0].corstats[icorstat].trace.data[irange])
        for i in range(1, len(st1)):
            signals = np.vstack((signals, st1[i].corstats[icorstat].trace.data 
                                 / np.linalg.norm(st1[i].corstats[icorstat].trace.data[irange])))
        # median and std
        n,m = signals.shape
        self.trace.std = np.empty(m)
        median = np.empty(m)
        for j in range(m):
            median_i, mad_std = _median_std(signals[:,j])
            median[j] = median_i
            self.trace.std[j] = mad_std
        # weighted mean + theoretical pick
        norm_median = np.linalg.norm(median)
        cormax = np.empty(n)
        p_theor = np.empty(n)
        for i in range(n):
            cormax[i] = np.amax([np.correlate(signals[i,:]/np.linalg.norm(signals[i,:]), median/norm_median, 
                                                                    mode='valid')[0], 0.])
            p_theor[i] = st1[i].stats.p_theor - st1[i].corstats[icorstat].trace.stats.starttime
        for j in range(m):
            weights = cormax.copy()
            for i in range(n):
                if np.absolute(signals[i,j]-median[j]) > 5 * self.trace.std[j]:
                    weights[i] = 0.
            self.trace.data[j] = np.average(signals[:,j], weights=weights)
        # relative time of the theoretical pick
        self.trace.stats.p_theor = np.median(p_theor)
        # other
        self.trace.stats.station = 'MEAN'
        self.trace.stats.network = ''
        return self
    
    def find_start(self, method="full", **kwargs):
        tm = self.trace.times()
        if method == 'trig':
            sta = kwargs['sta']
            lta = kwargs['lta']
            trig_thres = kwargs['trig_thres']
            p_aprox = kwargs['p_aprox']
            t_offset = [lta+6, 0]   # do not take initial part
            df = self.trace.stats.sampling_rate
            cft = recSTALTA(self.trace.data, int(sta * df), int(lta * df))
            t1 = self.trace.stats.starttime + t_offset[0]
            t2 = self.trace.stats.endtime + t_offset[1]
            max_cft, max_index, max_time = _find_max(cft, t1, t2, self.trace)
            trigs = triggerOnset(cft, trig_thres[0], trig_thres[1])
            if len(trigs) > 0:
                it1 = np.where(tm <= t_offset[0]-2)[0][-1]   # do not take initial part
                itrigs = np.where(trigs[:,0] >= it1)[0]
                if len(itrigs) > 0:
                    dtrigs = tm[trigs[itrigs, 0]] - p_aprox
                    itrigs2 = np.where(dtrigs >= 0)[0]
                    if len(itrigs2) > 0:
                        ii = trigs[itrigs[itrigs2[0]], 0]
                    else:
                        ii = trigs[itrigs[-1], 0]
                else:
                    trigs = []
            if len(trigs) == 0:
                ii = np.where((cft < trig_thres[0]) & (tm < (max_time - self.trace.stats.starttime)))[0]
                if ii.size > 0:
                    ii = ii[-1] + 1
                else:
                    ii = max_index
            self.wave_start = tm[ii]
        elif method == 'full_rough':
            # it needs roots already defined
            t_offset = [-10,0]  # from 0. root = 1. root - (2. root - 1. root)
            if len(self.roots) >= 2:
                cft_root0 = 2 * self.roots[0] - self.roots[1]
            elif len(self.roots) == 1:
                cft_root0 = self.roots[0] - 0.5
            else:
                cft_root0 = self.wave_start - 0.5
            t1 = self.trace.stats.starttime + (cft_root0 + t_offset[0])
            t2 = self.trace.stats.starttime + (cft_root0 + t_offset[1])
            noise = self.trace.slice(t1, t2)
            median, med_std = _median_std(noise.data)
            ii = np.where((np.absolute(self.trace.data - median) <= med_std) & 
                    #      (tm <= cft_root0 + 0.3*(self.roots[0] - cft_root0)))[0][-1]
                          (tm <= min([self.wave_start, self.maxmins[0]])))[0][-1]
            self.noise_level = [med_std, median, cft_root0 + t_offset[0], cft_root0 + t_offset[1]]
            self.wave_start = tm[ii]
        elif method == 'full_soft':
            # it needs trig start
            noise_offset = 10
            root_offset = 2
            peak_max = kwargs['peak_max']
            
            signal = self.trace.data
            i_root_offset = int(np.round(root_offset * self.trace.stats.sampling_rate))
            # find all roots
            cft = signal
            cft_sign = np.sign(cft)
            cft_diff = np.diff(cft_sign)
            ii = np.where(cft_diff != 0)[0]
            i1 = 0
            cft_roots, cft_roots_i = [], []
            for i2 in ii:
                if tm[i2]-tm[i1] < 0.25: continue
                pom = np.absolute(cft[i2:i2+3])
                j = max([0, i2 + np.where(pom == min(pom))[0][0] - 1])
                cft_roots_i.append(j)
                cft_roots.append(tm[j])
                i1 = i2
            cft_roots_all = np.array(cft_roots)
            cft_roots_all_i = np.array(cft_roots_i)
            # peak loop
            jj = np.where(cft_roots_all >= self.wave_start)[0]
            if len(jj) > 0:
                j2 = jj[0]
            else:
                j2 = len(cft_roots_all)
            ii = -1
            while j2 >= 0:
                # maxmin
                i2 = cft_roots_all_i[j2]
                if j2 >= 2:
                    i1 = max([cft_roots_all_i[j2-2], i2 - i_root_offset])
                else:
                    i1 = 0
                i_maxmin = i1 + np.argmax(np.absolute(signal[i1:i2]))
                t_maxmin = tm[i_maxmin]
                maxmin = signal[i_maxmin]
                # noise
                cft_root0 = tm[i1]
                t1 = self.trace.stats.starttime + (cft_root0 - noise_offset)
                t2 = self.trace.stats.starttime + cft_root0
                noise = self.trace.slice(t1, t2)
                median, med_std = _median_std(noise.data)
                # wave start
                if np.absolute(maxmin - median) <= 3 * med_std:
                    # noise
                    cft_root0 = tm[i2]
                    t1 = self.trace.stats.starttime + (cft_root0 - noise_offset)
                    t2 = self.trace.stats.starttime + cft_root0
                    noise = self.trace.slice(t1, t2)
                    median, med_std = _median_std(noise.data)
                    ii = np.where((np.absolute(signal - median) >= med_std) & 
                          (tm > cft_roots_all[j2]))[0]
                    if len(ii) > 0:
                        ii = ii[0]
                    else:
                        ii = -1
                        break
                    self.noise_level = [med_std, median, cft_root0 - noise_offset, cft_root0]
                    # roots
                    j2 = np.where(cft_roots_all > tm[ii])[0][0]
                    j3 = min([j2+peak_max, len(cft_roots_all)])
                    self.roots = cft_roots_all[j2:j3]
                    # maxmins
                    cft_maxmins = []
                    for j in range(j2-1, j3-1):
                        i1 = cft_roots_i[j]
                        i2 = cft_roots_i[j+1]
                        i_maxmin = i1 + np.argmax(np.absolute(signal[i1:i2]))
                        if tm[i_maxmin] > tm[ii]:
                            cft_maxmins.append(tm[i_maxmin])
                    self.maxmins = np.array(cft_maxmins)
                    break
                j2 -= 1
            if ii >= 0:
                self.wave_start = tm[ii]
            else:
                self.wave_start = 1e10
        elif method == 'full':
            # one of full_rough and full_soft closer to p_theor
            peak_max = kwargs['peak_max']
            pick0 = self.wave_start
            self.find_start(method='full_rough')
            pick1 = self.wave_start
            self.find_start(method='full_soft', peak_max=peak_max)
            pick2 = self.wave_start
            if np.absolute(pick1 - self.trace.stats.p_theor) <= np.absolute(pick2 - self.trace.stats.p_theor):
                # once again (to have correct roots and noise)
                self.wave_start = pick0
                self.find_maxmin(peak_max=peak_max)
                self.find_start(method='full_rough')
            else:
                self.wave_start = pick2
        return self
    
    def find_maxmin(self, peak_max, t_offset=-1):
        # PARAMETERS:
        #bb_period_min = 0.1  # bandpass filter to find local roots - OBSOLETE
        #bb_period_max = 4.0
        #t_offset = -1        # how far from trig_start the maxmin can be found
        
        # find roots first
        tr_median_filt = self.trace.copy()
        #tr_median_filt.filter('bandpass', freqmin=1/bb_period_max, freqmax=1/bb_period_min, corners=2, zerophase=True)
        tm = tr_median_filt.times()
        
        cft = tr_median_filt.data
        cft_sign = np.sign(cft)
        cft_diff = np.diff(cft_sign)
        ii = np.where(cft_diff != 0)[0]
        i1 = 0
        cft_roots, cft_roots_i = [], []
        for i2 in ii:
            if tm[i2]-tm[i1] < 0.25: continue
            pom = np.absolute(cft[i2:i2+3])
            j = max([0, i2 + np.where(pom == min(pom))[0][0] - 1])
            cft_roots_i.append(j)
            cft_roots.append(tm[j])
            i1 = i2
        cft_roots_all = np.array(cft_roots)
        cft_roots_all_i = np.array(cft_roots_i)
        ii_roots = np.where(cft_roots_all >= self.wave_start)[0]
        if len(ii_roots) == 0: return self
        cft_roots = cft_roots_all[ii_roots[0]:ii_roots[0]+peak_max]
        cft_roots_i = cft_roots_all_i[ii_roots[0]:ii_roots[0]+peak_max]

        # find maxmins
        cft_maxmins = []
        corr_median = self.trace.data
        i1 = np.where(tm <= self.wave_start + t_offset)[0][-1]
        for i2 in cft_roots_i:
            if i1 != i2:
                ii = i1 + np.argmax(np.absolute(corr_median[i1:i2]
                                - 0.5 * (corr_median[i1] + corr_median[i2])))
                cft_maxmins.append(tm[ii])
            i1 = i2
        
        # forgotten roots
        ii = np.where(cft_roots_all[:ii_roots[0]] > cft_maxmins[0])[0]
        if len(ii) > 0:
            cft_roots = cft_roots_all[ii[0]:ii[0]+peak_max]
            cft_roots_i = cft_roots_all_i[ii[0]:ii[0]+peak_max]
            # find maxmins once more
            cft_maxmins = []
            corr_median = self.trace.data
            i1 = np.where(tm <= self.wave_start + t_offset)[0][-1]
            for i2 in cft_roots_i:
                ii = i1 + np.argmax(np.absolute(corr_median[i1:i2]
                                    - 0.5 * (corr_median[i1] + corr_median[i2])))
                cft_maxmins.append(tm[ii])
                i1 = i2
            
        self.roots = np.array(cft_roots)
        self.maxmins = np.array(cft_maxmins)
        return self
    
    def set_noise_level(self, wave_start):
        # median of analytical signal envelope
        t_offset = [-5,0]   # noise signal, from wave_start
        tm = self.trace.times()
        ii = np.where((tm >= wave_start + t_offset[0]) & (tm < wave_start + t_offset[1]))[0]
        # noise #1 - envelope
        noise_env = envelope(self.trace.data)
        median, med_std = _median_std(noise_env[ii])
        # noise #2 - maxmin
        imax = np.argmax(self.trace.data[ii])
        imax = ii[imax]
        noise_max = self.trace.data[imax]
        weight_max = 0.5 * (tm[imax] - wave_start - t_offset[0]) / (t_offset[1] - t_offset[0])
        imin = np.argmin(self.trace.data[ii])
        imin = ii[imin]
        noise_min = self.trace.data[imin]
        weight_min = 0.5 * (tm[imin] - wave_start - t_offset[0]) / (t_offset[1] - t_offset[0])
        # combination of noise #1 and #2
        noise1 = weight_min * noise_min - (1 - weight_min) * median
        noise2 = weight_max * noise_max + (1 - weight_max) * median
        noise_mean = 0.5 * (noise1 + noise2)
        noise_std = noise2 - noise_mean
        self.noise_level = [noise_std, noise_mean, wave_start + t_offset[0], wave_start + t_offset[1]]
        return self

class Picks(object):
    def __init__(self):
        self.p_maxim = np.zeros(3)
        self.p_parabol = np.zeros(3)
        self.p_top = np.zeros(3)
        self.p_top_range = []
        self.p_correl = np.zeros(3)
        self.p_final = np.zeros(2)
    def set_picks(self, method, irange, trace, **kwargs):
        # kwargs: maxim_sign, noise, (p_correl)
        tm = trace.times()
        if method == 'maxim':       # [time, maxmin, SNR]
            if 'maxim_sign' in kwargs:
                if kwargs['maxim_sign'] > 0:
                    ii = np.argmax(trace.data[irange])
                else:
                    ii = np.argmin(trace.data[irange])
            else:
                ii = np.argmax(np.absolute(trace.data[irange]))
            imax = irange[ii]
            # expand if on boundaries
            if ii == 0:
                i1 = 0; i2 = irange[2]
                ik = -1
            elif ii == len(irange):
                i1 = irange[-2]; i2 = len(trace.data)
                ik = 0
            if (ii == 0) or (ii == len(irange)):
                signal = trace.data[i1:i2]
                ds = np.sign(np.diff(signal))
                if ds[ik] != 0:
                    imax = np.where(ds == -ds[ik])[0][ik] + i1 + 1
            self.p_maxim[0] = tm[imax]
            self.p_maxim[1] = trace.data[imax]
            self.p_maxim[2] = abs(trace.data[imax] / kwargs['noise'])   # signal_amplitude / noise_std
        elif method == 'parabol':   # [time, maxmin, width]
            c_top_part = 0.75  # take only most upper part near maximum
            signal = trace.data[irange]
            top_part = np.sort(np.array([c_top_part * self.p_maxim[1], self.p_maxim[1]]))
            ii = np.where((signal >= top_part[0]) & (signal <= top_part[1]))[0]
            irange = irange[ii]
            signal = trace.data[irange]
            time = tm[irange]
            coef = np.polyfit(time, signal, 2)
            self.p_parabol[0] = -0.5 * coef[1] / coef[0]
            self.p_parabol[1] = coef[2] - 0.25 * coef[1]**2 / coef[0]
            self.p_parabol[2] = 0.5 * np.sqrt(coef[1]**2 - 4 * coef[0] * (coef[2] - 0.5 * self.p_parabol[1])) \
                                / abs(coef[0])
        elif method == 'top':       # [time, maxmin, width], irange
            # it uses a noise level to mark top area of the extrem where a pick should be found in case of noisy signal
            noise = kwargs['noise']
            # expand irange if p_maxim is out of it
            ipmax = np.where(tm >= self.p_maxim[0])[0][0]
            if ipmax < irange[0]:
                irange = range(ipmax, irange[-1]+1)
            elif ipmax > irange[-1]:
                irange = range(irange[0], ipmax+1)
            signal = trace.data[irange]
            ii = np.where(noise >= np.absolute(self.p_maxim[1] - signal))[0]
            iid = np.where(np.diff(ii) != 1)[0]
            if iid.size > 0:
                starts = np.insert(ii[iid+1], 0, ii[0])
                ends = np.append(ii[iid], ii[-1])
                mins = np.zeros(len(starts))
                for i,(i1,i2) in enumerate(zip(starts,ends)):
                    mins[i] = np.amin(np.absolute(tm[irange[i1:i2+1]]-self.p_maxim[0]))
                i = np.argmin(mins)
                i1 = starts[i]; i2 = ends[i]
            else:
                i1 = ii[0]; i2 = ii[-1]
            irange1 = irange[i1]
            irange2 = irange[i2]
            # expand boundaries if neccessary
            if i1 == 0:
                signal = trace.data[:irange[0]]
                ii = np.where(noise < np.absolute(self.p_maxim[1] - signal))[0]
                if ii.size > 0: irange1 = ii[-1] + 1
            if i2 == len(irange)-1:
                signal = trace.data[irange[-1]+1:]
                ii = np.where(noise < np.absolute(self.p_maxim[1] - signal))[0]
                if ii.size > 0: irange2 = ii[0] + irange[-1]
            irange0 = int(round(0.5 * (irange1 + irange2)))
            self.p_top[0] = tm[irange0]
            self.p_top[1] = trace.data[irange0]
            self.p_top[2] = tm[irange2] - tm[irange1]
            self.p_top_range = range(irange1,irange2+1)
        elif method == 'final':    # [time, width]
            # combination of Picks: maxim + correl(global) + correl(local)
            p_correl_global = kwargs['p_correl_global']
            noise = kwargs['noise']
            i_peak = kwargs['peak'] + 1
            picks = np.zeros(3)
            stds = np.zeros(3)
            # maxim
            picks[0] = self.p_maxim[0]
            stds[0] = 0.5 * self.p_top[2] + 0.5 * (1-p_correl_global[1]) * i_peak
            # global correl (one period)
            picks[1] = p_correl_global[0]
            stds[1] = 0.5 * p_correl_global[2] + 0.5 * (1-p_correl_global[1]) * i_peak
            # local correl (half period)
            picks[2] = self.p_correl[0]
            stds[2] = 0.5 * self.p_correl[2]/2 + 0.5 * (1-self.p_correl[1]) + \
                        0.5 * (1-p_correl_global[1]) * (i_peak-1)
            coef = fit_gaussian(picks, stds)
            self.p_final[0] = coef[0]
            self.p_final[1] = 2 * coef[1]
            # low SNR -> large deviation
            snr = np.absolute(self.p_maxim[1]) / noise
            std_noise = max([2. - snr, 0.])
            self.p_final[1] += std_noise
        return self

#--- FUNCTIONS ----------------------------------------------

def _find_max(charfct, starttime, endtime, trace):
    
    # find maximum of charfct in given interval
    starttime0 = trace.stats.starttime
    delta = trace.stats.delta
    npts = trace.stats.npts
    i_start = int(round((starttime - starttime0) / delta))
    i_end = int(round((endtime - starttime0) / delta))
    i_start = min(i_start, npts-1); i_start = max(i_start, 0)
    i_end = min(i_end, npts-1); i_end = max(i_end, 0)
    if i_start > i_end:
        i_start = 0
        i_end = npts - 1
    cft = charfct[i_start:i_end+1]
    max_icft = np.argmax(cft)
    max_cft = cft[max_icft]
    max_index = max_icft + i_start
    
    # expand if on interval edges
    if max_index == i_start and max_index > 0:
        n = max_index
        max_index = 0
        for i in range(n-1, -1, -1):
            if np.argmax(charfct[i:n+1]) != 0:
                max_index = i + 1
                break
    elif max_index == i_end and max_index < npts-1:
        n = max_index
        max_index = npts - 1
        for i in range(n+1, npts):
            if np.argmax(charfct[n:i+1]) != (i-n):
                max_index = i - 1
                break
    
    max_time = starttime0 + (max_index * delta)
    return max_cft, max_index, max_time

def quantile_1D(data, weights, quantile):
    """
    Compute the weighted quantile of a 1D numpy array.
    Parameters
    ----------
    data : ndarray
        Input array (one dimension).
    weights : ndarray
        Array with the weights of the same size of `data`.
    quantile : float
        Quantile to compute. It must have a value between 0 and 1.
    Returns
    -------
    quantile_1D : float
        The output value.
    """
    # Check the data
    if not isinstance(data, np.matrix) :
        data = np.asarray(data)
    if not isinstance(weights, np.matrix) :
        weights = np.asarray(weights)
    nd = data.ndim
    if nd != 1:
        raise TypeError("data must be a one dimensional array")
    ndw = weights.ndim
    if ndw != 1:
        raise TypeError("weights must be a one dimensional array")
    if data.shape != weights.shape:
        raise TypeError("the length of data and weights must be the same")
    if ((quantile > 1.) or (quantile < 0.)):
        raise ValueError("quantile must have a value between 0. and 1.")
    # Sort the data
    ind_sorted = np.argsort(data)
    sorted_data = data[ind_sorted]
    sorted_weights = weights[ind_sorted]
    # Compute the auxiliary arrays
    Sn = np.cumsum(sorted_weights)
    # TODO: Check that the weights do not sum zero
    Pn = (Sn-0.5*sorted_weights)/np.sum(sorted_weights)
    # Get the value of the weighted median
    return np.interp(quantile, Pn, sorted_data)

def _median_std(data):
    # No weights
    weights = (np.zeros(len(data)) + 1) / len(data)
    median = quantile_1D(data, weights, 0.5)
    #mad_std = 1.4826 * quantile_1D([abs(x-median) for x in data], weights, 0.5)
    mad_std = quantile_1D([abs(x-median) for x in data], weights, 0.68)
    return median, mad_std

def _correlate(tm1, signal1, t1, tm2, signal2, t2):
    # it should be t1[0] <= t2[0] < t2[1] <= t1[1]
    ii1 = np.where((tm1 >= t1[0]) & (tm1 <= t1[1]))[0]
    ii2 = np.where((tm2 >= t2[0]) & (tm2 <= t2[1]))[0]
    sig1 = signal1[ii1] / np.linalg.norm(signal1[ii1])
    sig2 = signal2[ii2] / np.linalg.norm(signal2[ii2])
    correl = np.correlate(sig1, sig2, mode='valid')
    imax = np.argmax(correl)
    cor_shift = t2[0] - tm1[ii1[imax]]
    ii1 = ii1[imax:imax+len(ii2)]
    sig1 = signal1[ii1] / np.linalg.norm(signal1[ii1])
    cor_max = np.correlate(sig1, sig2, mode='valid')[0]
    return cor_max, cor_shift

def _draw_pick(ax, time, text='', color='k', zorder=1):
    koef = 0.8  # how big part of y axis
    ylims = ax.get_ylim()
    plims = [0.5*(ylims[0]+ylims[1]-koef*(ylims[1]-ylims[0])),0.5*(ylims[0]+ylims[1]+koef*(ylims[1]-ylims[0]))]
    ax.plot([time,time], plims, color=color, zorder=zorder)
    ax.text(time+0.1, plims[1], text, color=color)
    
def dgauss(x, *p):
    mu, sigma = p
    return np.exp(-(x-mu)**2/(2.*sigma**2))/np.sqrt(2.*np.pi)/sigma

def fit_gaussian(means, stds, draw=False):
    """
    Final probability distribution is considered as a stack of the Gaussian distributions given by means and stds:
    S(x) = sum_i [1/(sqrt(pi)*stds_i) * exp(-0.5*((x-means_i)/stds_i)^2)]
    Result: (ampl, mean, std) for the Gaussian which fits the final distribution S(x)
    """    
    means = np.array(means)
    stds = np.array(stds)
    if draw:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    # find input x-interval
    dx = min(stds)/10.
    x1 = min([mean-5*std for mean,std in zip(means,stds)])
    x2 = max([mean+5*std for mean,std in zip(means,stds)])
    x = np.arange(x1, x2+dx, dx)
    # final distribution
    Sy = np.zeros(len(x))
    for mean,std in zip(means,stds):
        y = dgauss(x, *[mean, std])
        Sy += y
        if draw: ax.plot(x,y,'gray')
    #Sy = Sy / np.sum(Sy) / dx
    Sy = Sy / len(means)    # this should normalized Sy better
    if draw: ax.plot(x,Sy,'k')
    # initial guess
    Smean = np.average(means)
    Sstd = np.sqrt(np.average(stds**2 + (means-Smean)**2))
    y = dgauss(x, *[Smean, Sstd])
    if draw: ax.plot(x,dgauss(x, *[Smean, Sstd]),'b')
    # fit by the Gaussian - try several times with different starting inputs if curve_fit does not converge
    niter = 1000
    mmax = np.amax(means); mmin = np.amin(means); mm = 0.5 * (mmax - mmin); mmax += mm; mmax -= mm
    Smeans = np.append(Smean, means); 
    Smeans = np.append(Smeans, np.random.rand(niter) * (mmax-mmin) + mmin)
    mmax = np.amax(stds); mmin = np.amin(stds); mm = 0.5 * (mmax - mmin); mmax += mm; mmax -= mm
    Sstds = np.append(Sstd, stds); 
    Sstds = np.append(Sstds, np.random.rand(niter) * (mmax-mmin) + mmin)
    for p0 in zip(Smeans,Sstds):
        try:
            coeff, var_matrix = curve_fit(dgauss, x, Sy, p0=p0)
            break
        except Exception:
            coeff = [Smean, Sstd]
    if draw:
        y = dgauss(x, *coeff)
        ax.plot(x,y,'r')
        ax.set_xlim([x[0],x[-1]])
        plt.show()
    #print('{:6.2f} +- {:.2f}'.format(Smean, Sstd))
    return coeff

def _theo_arrivals(st, origin):
    # Theoretical arrivals for a center point of the array
    n = len(st)
    # find the center point (aproximate method)
    lats = np.zeros(n)
    lons = np.zeros(n)
    for i,tr in enumerate(st):
        lats[i] = tr.stats.latitude
        lons[i] = tr.stats.longitude
    lat0 = np.average(lats)
    lon0 = np.average(lons)
    # arrivals for the center point
    _model = TauPyModel(model=u'iasp91')
    [dist, az, baz] = dist_baz(lat0, lon0, origin.latitude, origin.longitude)
    # fix error in old taup
    depth_fix = 1.25
    depth = origin.depth / 1000.
    if (depth > 0) and (depth < depth_fix):
        if depth > depth_fix-depth:
            depth = 0.
        else:
            depth = depth_fix
    arrivals = _model.get_travel_times(source_depth_in_km=depth, distance_in_degree=dist)
    return arrivals

def _mseed_name(net, stat, loc, chan, comp, origin, ext='.P'):
    return '{}.{}.{}.{}{}.{}{}'.format(net, stat, loc, chan, comp, origin.strftime("%y%m%d%H%M%S"), ext)

def _str_event(event):
    origin = event.preferred_origin() or event.origins[0]
    magnitude = event.preferred_magnitude() or event.magnitudes[0]
    # healing
    if origin.depth is None:
        origin.depth = 0.
    try:
        magn_author = magnitude.creation_info.author.upper()
    except Exception:
        magn_author = 'unk.'
    try:
        origin_author = origin.creation_info.author.upper()
    except Exception:
        origin_author = 'unk.'
    sss = "{:s} {:7.2f} {:7.2f} {:4.0f} {:4.1f}{:<3s}  {:s} ({:s}/{:s})".format(
            origin.time.strftime("%y/%m/%d  %H:%M:%S.%f")[:-5], origin.latitude, origin.longitude,
            origin.depth/1000., magnitude.mag, magnitude.magnitude_type,
            event.event_descriptions[0].text, origin_author, magn_author)
    return sss

def _round_time(time, sample_rate=100., ceil=False):
    nprec = int(np.ceil(np.log10(sample_rate)))
    nprec = min([max([0, nprec]), 5])
    time_new = deepcopy(time)
    ms = time.microsecond
    ms_new = np.around(ms, decimals=nprec-6)
    if ms_new >= 1000000:
        time_new += 1
        time_new.microsecond = 0
    else:
        time_new.microsecond = ms_new
    test = ms_new != ms
    if ceil and (ms_new < ms):  
        time_new += 1./sample_rate   # we need ceil for interpolation
    return time_new, test

def _find_in_list(element, list_element):
    try:
        index_element = list_element.index(element)
        return index_element
    except ValueError:
        return -1

def _read_and_prepare_mseed(fold_archive, net, stat, loc, origin, start, end, routefile, parser, inv, 
                            sample_rate, npts, only_archive=False):
    # only Z component
    comp = 'Z'
    st = Stream()
    inv2 = inv.select(network=net, station=stat, location=loc, channel="[HBES]HZ", starttime=start, endtime=end)
    if not inv2.networks:
        print('!  {}.{}.{}.?H{} | {} - {} | not in inventory'.format(
                net, stat, loc, comp, start.isoformat(), end.isoformat()))
        return st
    channels = []
    for channel in inv2.networks[0].stations[0].channels:
        channels.append(channel.code[0:2])
    if not channels:
        print('! no Z component in:')
        print(inv)
        return st
    channels_old = list(set(channels))
    ibh = _find_in_list('BH', channels_old)
    ihh = _find_in_list('HH', channels_old)
    ish = _find_in_list('SH', channels_old)
    ieh = _find_in_list('EH', channels_old)
    channels = list(channels_old)
    if (ibh>=0) and (ihh>=0):
        channels.remove('BH')
        channels.remove('HH')
        channels.append('BB')
    if (ish>=0) and (ieh>=0):
        channels.remove('SH')
        channels.remove('EH')
        channels.append('SP')
    # read mseed
    for chan1 in channels:
        if chan1 in ['BB', 'BH', 'HH']: chan2 = 'HH'
        if chan1 in ['SP', 'SH', 'EH']: chan2 = 'EH'
        chan2 = '[' + chan2[0].upper() + chan2[0].lower() + ']' + chan2[1]
        origin_num = origin.strftime("%y%m%d%H%M%S")
        fold = os.path.join(fold_archive, origin_num)
        if not os.path.isdir(fold):
            os.makedirs(fold)
        phase1 = phase
        if (phase == 'pP') or (phase == 'sP') or (phase == 'PcP'): phase1 = 'P'   # use same mseed files
        mseed = os.path.join(fold, _mseed_name(net, stat, loc, chan2, comp, origin, '.'+phase1))
        try:
            st += read(mseed, format='MSEED')
            for tr in st: print(tr)
        except Exception:
            if not only_archive:
                st1 = read_mseed_route(network=net, station=stat, location=loc, channel=chan1, component=comp, 
                                starttime=start, endtime=end, verbose=1, routefile = routefile, check_all_archives=False)
                st1 = better_merge(st1, edge_gap=0.1)
                # remove bad traces
                delta0 = end - start
                for tr in st1:
                    # not enough data
                    delta = tr.stats.endtime - tr.stats.starttime
                    if (delta0-delta)/delta0 > 0.1:
                        print('!  {} not complete, removed'.format(tr.id))
                        st1.remove(tr)
                        continue
                    # zero signal
                    absmax = np.amax(np.absolute(tr.data))
                    if absmax == 0.:
                        print('!  {} zeros only, removed'.format(tr.id))
                        st1.remove(tr)
                # wwssn_sp simulation
                st1 = simul_wwssn_sp(st1, parser, demedian=False, output="DISP")
                for tr in st1:
                    # sample rate & starttime check
                    start_old = tr.stats.starttime
                    start_new, io_start = _round_time(tr.stats.starttime, sample_rate=sample_rate, ceil=True)
                    chan = tr.stats.channel[:2]
                    sample_rate_old = tr.stats.sampling_rate
                    io_sample = sample_rate_old != sample_rate
                    if io_sample or io_start:
                        tr.data = np.append(tr.data, np.zeros(int(0.2 * delta0 * sample_rate_old)))
                        tr.data = np.insert(tr.data, 0, tr.data[0])
                        tr.stats.starttime = tr.stats.starttime - 1./sample_rate_old
                        tr.interpolate(starttime=start_new, sampling_rate=sample_rate, npts=npts)
                        if io_sample:
                            chan = chan.replace('SH','EH').replace('BH','HH')
                            tr.stats.channel = chan + tr.stats.channel[2]
                            if sample_rate_old < sample_rate:
                                chan = chan[0].lower() + chan[1]
                        if io_sample: print('   resample {:.0f} Hz -> {:.0f} Hz'.format(sample_rate_old, sample_rate))
                        if io_start: print('   change start {} -> {}'.format(start_old.isoformat(), start_new.isoformat()))
                    # write it to archive folder
                    _name = _mseed_name(net=tr.stats.network, stat=tr.stats.station, loc=tr.stats.location, 
                                        chan=chan, comp=tr.stats.channel[2], origin=origin)
                    tr.write(os.path.join(fold, _name), format='MSEED')
                st += st1
                st.write(mseed, format='MSEED')
        # CORRECTIONS
        if (net == 'SL') and (stat == 'GBRS') and (comp == 'Z'):      # reversed polarity
            for tr in st:
                tr.data = -tr.data
        elif (net == 'BW') and (stat == 'MGBB') and \
             (origin >= UTCDateTime('2014-11-06')) and (origin <= UTCDateTime('2014-11-23')):
            st = Stream()
        elif (net == 'CZ') and (stat == 'PRU'):                       # CZ.PRU..EHZ too noisy
            for tr in st.select(channel='EHZ'):
                st.remove(tr)
    return st

#--- MAIN ---------------------------------------------------

# write header
print('! script: ' + os.path.basename(__file__))
print('! version: ' + __version__)
print('! event catalog: ' + catalog)
print('! run at ' + UTCDateTime().isoformat())

# read events
cat = readEvents(catalog)

for i_evt, event in enumerate(cat):

    origin = event.preferred_origin() or event.origins[0]
    id_origin = origin.time.strftime("%y%m%d%H%M%S")

    output = os.path.join(fold_out, 'pick_'+id_origin+'_'+phase+'.txt')
    if os.path.isfile(output):   # let start from not done yet
        continue

    print('\n#' + str(i_evt) + '  ' +_str_event(event)+'\n')

    fold_out_pics_evt = os.path.join(fold_out_pics, id_origin)
    if not os.path.isdir(fold_out_pics_evt):
        os.makedirs(fold_out_pics_evt)    

    stats = GRP

    #--- READ MSEEDS ------------
    st_all = Stream()
    for stat in stats:
        # read response
        inv = read_inventory(glob.glob(os.path.join(fold_dless, '??.' + stat + '.xml'))[0], format='STATIONXML')
        parser = Parser(glob.glob(os.path.join(fold_dless, '??.' + stat + '.seed'))[0])
        # station coordinates
        reflat = inv.networks[0].stations[0].latitude
        reflon = inv.networks[0].stations[0].longitude
        net = inv.networks[0].code
        # dist, baz, incid
        [dist, az, baz] = dist_baz(reflat, reflon, origin.latitude, origin.longitude)
        if origin.depth is None: origin.depth = 0
        [p_phase, p_offset, inci] = arrival_incid(origin.depth, dist, phases=phase, model=u'iasp91')
        if not p_phase:
            continue
        inci = inci[0]
        pk_date = origin.time + p_offset[0]
        # nacti mseed
        starttime = pk_date + mseed_offset[0]
        endtime = pk_date + mseed_offset[1]
        st = _read_and_prepare_mseed(fold_archive=fold_mseed, net=net, stat=stat, loc='*', origin=origin.time,
                        start=starttime, end=endtime, routefile=routefile, parser=parser, inv=inv,
                        sample_rate=sample_rate, npts=int((mseed_offset[1]-mseed_offset[0])*sample_rate)+1,
                        only_archive=io_only_archive)
        if not st:
            continue
        for tr in st:
            tr.stats.p_theor = pk_date
            tr.stats.latitude = reflat
            tr.stats.longitude = reflon
        st_all += st

    #--- TRIGGER ------------
    trig_max_loc, trig_max_glob = [], []
    trig_offs_loc, trig_offs_glob = [], []

    for tr in st_all:
        # trigger
        df = tr.stats.sampling_rate
        cft = recSTALTA(tr.data, int(sta * df), int(lta * df))
        # local max
        t1 = pk_date - 5
        t2 = pk_date + 10
        max1_cft, max1_index, max1_time = _find_max(cft, t1, t2, tr)
        trig_max_loc.append(max1_cft)
        trig_offs_loc.append(max1_time - pk_date)
        # global max
        t1 = starttime + 2 * lta
        t2 = endtime
        max2_cft, max2_index, max2_time = _find_max(cft, t1, t2, tr)
        trig_max_glob.append(max2_cft)
        trig_offs_glob.append(max2_time - pk_date)

        if io_figure and not io_only_archive:
            plotTrigger(tr, cft, trig_t1, trig_t2, show=False)
            # add max lines to figure
            fig = plt.gcf()
            ax = fig.gca()
            ax.axvline(max2_time - tr.stats.starttime, color='y')
            ax.axvline(max1_time - tr.stats.starttime, color='g')
            output_fig = os.path.join(fold_out_pics_evt, id_origin+'_'+tr.id+'_trig.png')
            fig.savefig(output_fig)
            plt.close(fig)

    trig_max_loc_median, trig_max_loc_std = _median_std(trig_max_loc)
    trig_offs_loc_median, trig_offs_loc_std = _median_std(trig_offs_loc)
    trig_max_glob_median, trig_max_glob_std = _median_std(trig_max_glob)
    trig_offs_glob_median, trig_offs_glob_std = _median_std(trig_offs_glob)

    print('\ntrig_max_local:  {0:.2f} +- {1:.2f}\ttrig_offs_local:  {2:.2f} +- {3:.2f}'.format(
            trig_max_loc_median, 2*trig_max_loc_std, trig_offs_loc_median, 2*trig_offs_loc_std))
    print('trig_max_global: {0:.2f} +- {1:.2f}\ttrig_offs_global: {2:.2f} +- {3:.2f}'.format(
            trig_max_glob_median, 2*trig_max_glob_std, trig_offs_glob_median, 2*trig_offs_glob_std))

    n_chybi = len(stats) - len(trig_max_loc)
    max_enh = trig_max_loc_median - n_chybi / 20
    if max_enh < 2.8:
        evt_take = False
    elif max_enh >= 3.5:
        evt_take = True
    else:
        if trig_offs_loc_std <= 2 * max_enh - 5:
            evt_take = True
        else:
            evt_take = False
    print('-> recommend to take event: {}\n'.format(evt_take))


    #--- LONG CORRELATION ------

    t_offsets = [-10, 5]          # from sta/lta energy maximum
    tcor_off = [-0.5, 0.5]        # correlation maximum search interval

    for tr in st_all:
        tr.corstats = []
        tr.corstats.append(Correl_stats())
        start_corr = tr.stats.p_theor + (trig_offs_loc_median + t_offsets[0])
        end_corr = tr.stats.p_theor + (trig_offs_loc_median + t_offsets[1])
        tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)

    # median from uncorrelated signals
    medstats = []
    medstats.append(Correl_stats())
    medstats[0].median(st_all, icorstat=0)    

    n_st = st_all.count()
    cor_max = np.empty([n_st,n_st])
    cor_shift = np.empty([n_st,n_st])

    # signal to signal correlations
    for i in range(n_st):
        cor_max[i,i] = 1.
        cor_shift[i,i] = 0.
        for j in range(i+1,n_st):
            tr1 = st_all[i].corstats[0].trace
            tr2 = st_all[j].corstats[0].trace
            tr1_signal = tr1.data / np.linalg.norm(tr1.data)
            tr2_signal = tr2.data / np.linalg.norm(tr2.data)
            n1 = tr1.stats.npts; n2 = tr2.stats.npts
            n = max(n1,n2); n2 = int(n/2); nmod = (n+1) % 2
            dt = tr1.stats.delta
            time = np.arange(-n2*dt, (n2-nmod)*dt+dt/2, dt)
            correl = np.correlate(tr1_signal, tr2_signal, mode='same')
            ii = np.where((time >= tcor_off[0]) & (time <= tcor_off[1]))[0]
            imax = np.argmax(correl[ii])
            imax = ii[imax]
            cor_max[i,j] = correl[imax]
            cor_shift[i,j] = -time[imax]
            cor_max[j,i] = cor_max[i,j]
            cor_shift[j,i] = -cor_shift[i,j]

    # signal time shifts
    if not cor_max.any(): cor_max.fill(1.)
    shifts = np.average(cor_shift, axis=0, weights=cor_max)
    shift0 = np.median(shifts)
    ii = np.where(np.absolute(shifts - shift0) > 2 * (tcor_off[1] - tcor_off[0]))[0]
    if ii: shifts[ii] = shift0
    shifts = np.around(shifts, decimals=2)

    # FIG1
    if io_figure:
        fig = plt.figure(figsize=(16,5))
        ax1 = fig.add_subplot(121)
        for tr in st_all:
            tr1 = tr.corstats[0].trace
            tm1 = tr1.times()
            tr1_signal = tr1.data / np.linalg.norm(tr1.data)
            ax1.plot(tm1, tr1_signal, label=tr1.id)
        ax1.plot(tm1, medstats[0].trace.data, 'k-')
        ax1.set_xlim((0, max(tm1)))
        #ax1.legend(loc='upper left')

    # correlated signals
    t_offsets = [-10, 0] # from shifted corstat.trace
    for i, tr in enumerate(st_all):
        start_corr = tr.corstats[0].trace.stats.starttime + shifts[i] + t_offsets[0]
        end_corr = tr.corstats[0].trace.stats.endtime + shifts[i] + t_offsets[1]
        tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
    # check if all traces have same length
    npts = np.array([tr.corstats[0].trace.stats.npts for tr in st_all])
    n_min = min(npts)
    nn = [i for i, n in enumerate(npts) if n > n_min]
    for i in nn:
        print(st_all[i].corstats[0].trace)
        st_all[i].corstats[0].trace.data = st_all[i].corstats[0].trace.data[:n_min]

    # median from correlated signals
    medstats[0].median(st_all, icorstat=0)

    # take different window if theor phase is too far
    t_shift = 5
    if medstats[0].trace.stats.p_theor > (medstats[0].trace.stats.endtime-medstats[0].trace.stats.starttime) - t_shift:
        for i, tr in enumerate(st_all):
            start_corr = tr.corstats[0].trace.stats.starttime + t_shift
            end_corr = tr.corstats[0].trace.stats.endtime + t_shift
            tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
        medstats[0].median(st_all, icorstat=0)

    # another correlation, with larger offset
    tcor_off = [-2, 2]
    tm = medstats[0].trace.times()
    irange = np.where(tm >= 10)[0]
    mean = medstats[0].trace.data[irange] / np.linalg.norm(medstats[0].trace.data[irange])
    n = len(mean); n2 = int(n/2); nmod = (n+1) % 2
    dt = medstats[0].trace.stats.delta
    time = np.arange(-n2*dt, (n2-nmod)*dt+dt/2, dt)
    for i,tr in enumerate(st_all):
        tr1 = tr.corstats[0].trace
        ### KNOWN TIME ISSUES
        tcor_off1 = tcor_off
        if tr1.stats.station == 'MGBB':
            tcor_off1 = [-10, 10]
        ###
        tr1_signal = tr1.data[irange] / np.linalg.norm(tr1.data[irange])
        cormax0 = np.correlate(tr1_signal, mean, mode='valid')[0]
        correl = np.correlate(tr1_signal, mean, mode='same')
        ii = np.where((time >= tcor_off1[0]) & (time <= tcor_off1[1]))[0]
        imax = np.argmax(correl[ii])
        imax = ii[imax]
        corshift = np.around(time[imax], decimals=2)
        cormax = correl[imax]
        print('{:16s} {:3.0f} {:5.2f} ->{:3.0f} {:5.2f}'.format(tr.id, 100*cormax0, shifts[i], 100*cormax, 
                                                                shifts[i]+corshift))
        # shift the traces
        start_corr = tr.corstats[0].trace.stats.starttime + corshift
        end_corr = tr.corstats[0].trace.stats.endtime + corshift
        tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
        shifts[i] += corshift
        # new correlation matrix (for clustering)
        tr1 = tr.corstats[0].trace
        tr1_signal = tr1.data[irange] / np.linalg.norm(tr1.data[irange])
        for j in range(i+1,n_st):
            tr2 = st_all[j].corstats[0].trace
            tr2_signal = tr2.data[irange] / np.linalg.norm(tr2.data[irange])
            correl = np.correlate(tr1_signal, tr2_signal, mode='valid')[0]
            cor_max[i,j] = correl
            cor_max[j,i] = cor_max[i,j]
    medstats[0].median(st_all, icorstat=0)

    # FIG2
    if io_figure:
        ax2 = fig.add_subplot(122)
        for tr in st_all:
            tr2 = tr.corstats[0].trace
            tm2 = tr2.times()
            tr2_signal = tr2.data / np.linalg.norm(tr2.data)
            ax2.plot(tm2, tr2_signal, label=tr2.id)
        ax2.plot(tm2, medstats[0].trace.data, 'k-')
        std_old = np.log10(np.absolute(np.divide(medstats[0].trace.std, medstats[0].trace.data)))
        ylims_old = [np.floor(np.nanmin(std_old)), np.ceil(np.nanmax(std_old))]
        ylims_new = ax2.get_ylim()
        std_new = (ylims_new[1]-ylims_new[0])/(ylims_old[1]-ylims_old[0]) * \
                    (std_old-ylims_old[0]) + ylims_new[0]
        ax2.plot(tm2, std_new, 'k:')
        ax2.set_xlim((10, max(tm2)))
        output_fig = os.path.join(fold_out_pics_evt, id_origin+'_longcor_all.png')
        fig.savefig(output_fig)
        plt.close(fig)

    # FIG3
    if io_figure:
        df = medstats[0].trace.stats.sampling_rate
        cft = recSTALTA(medstats[0].trace.data, int(sta * df), int(lta * df))
        plotTrigger(medstats[0].trace, cft, trig_t1, trig_t2, show=False)
        fig = plt.gcf()
        output_fig = os.path.join(fold_out_pics_evt, id_origin+'_longcor_trig.png')
        fig.savefig(output_fig)
        plt.close(fig)

    peak_max1 = peak_max
    
    # theo arrivals
    arrivals = _theo_arrivals(st_all, origin)

    # start of wave (triggered)
    medstats[0].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2], 
                           p_aprox=medstats[0].trace.stats.p_theor)

    # take different window if trig pick is too far
    t_shift = 5
    if medstats[0].wave_start > (medstats[0].trace.stats.endtime-medstats[0].trace.stats.starttime) - t_shift:
        for i, tr in enumerate(st_all):
            start_corr = tr.corstats[0].trace.stats.starttime + t_shift
            end_corr = tr.corstats[0].trace.stats.endtime + t_shift
            tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
        medstats[0].median(st_all, icorstat=0)
        medstats[0].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2], 
                           p_aprox=medstats[0].trace.stats.p_theor)

    # find roots and maxmins
    medstats[0].find_maxmin(peak_max1)

    # solve too far shifted signals
    while len(medstats[0].roots) < peak_max1:
        t_shift = (peak_max1 - len(medstats[0].roots) + 1) * 1.
        for tr in st_all:
            start_corr = tr.corstats[0].trace.stats.starttime + t_shift
            end_corr = tr.corstats[0].trace.stats.endtime + t_shift
            tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
        medstats[0].median(st_all, icorstat=0)
        medstats[0].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2], 
                               p_aprox=medstats[0].trace.stats.p_theor)
        medstats[0].find_maxmin(peak_max1)
    #start_trig = medstats[0].wave_start
    
    # start of wave (full)
    medstats[0].find_start(peak_max=peak_max1)

    #--- MORE PHASES LOOP ---

    # save copy
    st_all_arch0 = st_all.copy()
    medstat_arch0 = medstats[0].copy()

    theo_dif = medstat_arch0.wave_start - medstat_arch0.trace.stats.p_theor
    if theo_dif > max_theo_dif:
        nphase = 2
    else:
        nphase = 1
    iphase = nphase
    wave_start = medstat_arch0.wave_start

    print('\nPhase : {:d}/{:d}'.format(iphase, nphase))
    while iphase > 0:
        st_all_arch = st_all_arch0.copy()
        medstats = []; medstats.append(medstat_arch0.copy())
        peak_max1 = peak_max
        phase1 = phase
        if (iphase == 1) and (nphase == 2):
            t_shift = wave_start - (medstats[0].trace.stats.endtime - medstats[0].trace.stats.starttime)
            for tr in st_all_arch:
                start_corr = tr.corstats[0].trace.stats.starttime + t_shift
                end_corr = tr.corstats[0].trace.stats.endtime + t_shift
                tr.corstats[0].trace = better_slice(tr, start_corr, end_corr)
            medstats[0].median(st_all_arch, icorstat=0)
            medstats[0].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2], 
                                   p_aprox=medstats[0].trace.stats.p_theor, io_figure=True)
            try:
                medstats[0].find_maxmin(peak_max1)
                medstats[0].find_start(peak_max=peak_max1)
            except Exception as e:
                print(e)
                iphase -= 1
                continue
        elif (iphase == 2) and (nphase == 2):
            phase1 = phase + 'X'

        # save copy
        medstat_arch = medstats[0].copy()
        peak_max1 = min([peak_max1, len(medstat_arch.roots)])

        #--- LOOPS ------

        # LOOP #1 over number of peaks
        # LOOP #2 for better fitting

        t_offset = [-0.5, 0.5]  # zoom around wave_start to find max_correl
        max_loop = 10  # for LOOP #2

        st_all = st_all_arch.copy()
        medstats = []
        for tr in st_all:
            tr.corstats = []

        print('')
        io_converg = np.zeros(peak_max1-peak_min+1, dtype=bool)
        wave_start_shifts = np.zeros(peak_max-peak_min+1)
        for ics, max_maxmin in enumerate(range(peak_min,peak_max1+1)):
            print('{:d} peaks:'.format(max_maxmin))
            medstats.append(medstat_arch.copy())

            for icount in range(max_loop):
                if len(medstats[ics].roots) < max_maxmin:
                    print('! number of roots lower than peak_max: {:d} vs {:d}'.format(len(medstats[ics].roots), max_maxmin))
                    break
                t21 = medstats[ics].wave_start
                t22 = medstats[ics].roots[max_maxmin-1]
                t11 = t21 + t_offset[0]
                t12 = t22 + t_offset[1]
                tm = medstats[ics].trace.times()
                shift_sum = 0.
                for i, tr in enumerate(st_all):
                    tr.corstats.append(st_all_arch[i].corstats[0].copy())
                    cormax, cor_shift = _correlate(tm, tr.corstats[ics].trace.data, [t11, t12], 
                                                    tm, medstats[ics].trace.data, [t21, t22])
                    shift_sum += abs(cor_shift)

                    # shift signals according to correlation
                    start_corr = tr.corstats[ics].trace.stats.starttime - cor_shift
                    end_corr = tr.corstats[ics].trace.stats.endtime - cor_shift
                    tr.corstats[ics].trace = better_slice(tr, start_corr, end_corr)

                # median from correlated signals
                ii = np.where((tm >= medstats[ics].wave_start) & (tm <= medstats[ics].roots[max_maxmin-1]))[0]
                medstats[ics].median(st_all, icorstat=ics, irange=ii)
                # start of wave (triggered)
                medstats[ics].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2], 
                                         p_aprox=medstats[ics].wave_start)
                # find roots and maxmins
                try:
                    medstats[ics].find_maxmin(peak_max1)
                except Exception as e:
                    print(e)
                    if (iphase==1) and (nphase==2):
                        break
                    else:
                        medstats[ics].roots = []
                # solve too far shifted signals
                while (len(medstats[ics].roots) < max_maxmin) and ((iphase!=1) or (nphase!=2)):
                    t_shift = (max_maxmin - len(medstats[ics].roots) + 1) * 1.
                    wave_start_shifts[ics] += t_shift
                    for tr in st_all:
                        start_corr = tr.corstats[ics].trace.stats.starttime + t_shift
                        end_corr = tr.corstats[ics].trace.stats.endtime + t_shift
                        tr.corstats[ics].trace = better_slice(tr, start_corr, end_corr)
                    ii = np.where(tm >= medstats[ics].wave_start)[0]
                    medstats[ics].median(st_all, icorstat=ics, irange=ii)
                    medstats[ics].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2],
                                            p_aprox=medstats[ics].wave_start)
                    medstats[ics].find_maxmin(peak_max1)
                    if len(medstats[ics].roots) == max_maxmin:
                        ii = np.where((tm >= medstats[ics].wave_start) & (tm <= medstats[ics].roots[max_maxmin-1]))[0]
                        medstats[ics].median(st_all, icorstat=ics, irange=ii)
                        medstats[ics].find_start(method='trig', sta=sta, lta=lta, trig_thres=[trig_t1, trig_t2],
                                                p_aprox=medstats[ics].wave_start)
                        medstats[ics].find_maxmin(peak_max1)
                # start of wave (full)
                medstats[ics].find_start(peak_max=peak_max1)

                print('- {:d} {:.2f}'.format(icount+1, shift_sum))
                if shift_sum <= 2 * n_st * tr1.stats.delta:
                    io_converg[ics] = True
                    break

            max_peaks = min([len(medstats[ics].roots), max_maxmin])
            # short correlation maxs
            t21 = medstats[ics].wave_start
            t22 = medstats[ics].roots[max_peaks-1]
            t11 = t21 + t_offset[0]
            t12 = t22 + t_offset[1]
            tm = medstats[ics].trace.times()
            for tr in st_all:
                cormax, cor_shift = _correlate(tm, tr.corstats[ics].trace.data, [t11, t12], 
                                                tm, medstats[ics].trace.data, [t21, t22])
                tr.corstats[ics].cor_max.append(cormax)

            # noise level
            medstats[ics].set_noise_level(medstats[ics].wave_start)
            for tr in st_all:
                tr.corstats[ics].set_noise_level(medstats[ics].wave_start)

            # ------- PICKS ------------------------------------------------------------------------------
            t_offset = [-0.5, 0.5]  # zoom around wave_start to find max_correl
            tm = medstats[ics].trace.times()

            # PICKS:
            # maxmin picky
            # korelacni picky, koreluje se vzdy dana pulperioda
            # top picky, maxmin se vyplni 'vodou' o vysce dane sumem pred signalem
            # OBSOLETE: fitovani parabolou
            # OBSOLETE: korelacni picky = p_maxim na median a prislusne korelovany pick na signalech

            roots = medstats[ics].roots
            roots = np.insert(roots, 0, medstats[ics].wave_start)

            # init
            for i in range(max_peaks):
                medstats[ics].picks.append(Picks())
                for tr in st_all:
                    tr.corstats[ics].picks.append(Picks())

            # picks
            for i in range(max_peaks):
                ii = np.where((tm >= roots[i]) & (tm <= roots[i+1]))[0]
                medstats[ics].picks[i].set_picks(method='maxim', irange=ii, trace=medstats[ics].trace,
                                          noise = medstats[ics].noise_level[0])
                medstats[ics].picks[i].set_picks(method='top', irange=ii, trace=medstats[ics].trace, 
                                                 noise=medstats[ics].noise_level[0])
                t21 = roots[i]
                t22 = roots[i+1]
                t11 = t21 + t_offset[0]
                t12 = t22 + t_offset[1]
                for tr in st_all:
                    tr.corstats[ics].picks[i].set_picks(method='maxim', irange=ii, trace=tr.corstats[ics].trace,
                                noise=tr.corstats[ics].noise_level[0], maxim_sign = np.sign(medstats[ics].picks[i].p_maxim[1]))          
                    tr.corstats[ics].picks[i].set_picks(method='top', irange=ii, trace=tr.corstats[ics].trace, 
                                                        noise=tr.corstats[ics].noise_level[0])
                    cormax, cor_shift = _correlate(tm, tr.corstats[ics].trace.data, [t11, t12], 
                                                    tm, medstats[ics].trace.data, [t21, t22])
                    tr.corstats[ics].picks[i].p_correl[0] = medstats[ics].picks[i].p_maxim[0] - cor_shift
                    tr.corstats[ics].picks[i].p_correl[1] = cormax
                    tr.corstats[ics].picks[i].p_correl[2] = medstats[ics].picks[i].p_top[2]
                    tr.corstats[ics].picks[i].set_picks(method='final', irange=ii, trace=tr.corstats[ics].trace, 
                                noise=tr.corstats[ics].noise_level[0], peak=i,
                                p_correl_global=[medstats[ics].picks[i].p_maxim[0], tr.corstats[ics].cor_max[-1], 
                                                 medstats[ics].picks[i].p_top[2]])

        # how much picks per signal, take best
        ibest = [-1, 0, 0., 1000.]
        nst = len(st_all)
        edge = 0.01
        ordinal = lambda n: "%d%s" % (n,"tsnrhtdd"[(n/10%10!=1)*(n%10<4)*n%10::4])

        print('')
        if not io_converg.any(): io_converg.fill(True)
        for ics, max_maxmin in enumerate(range(peak_min,peak_max1+1)):
            if (len(medstats[ics].roots) < max_maxmin) or not io_converg[ics]:
                continue
            sss = '{:d} peaks:'.format(max_maxmin)
            for ip in range(max_maxmin):
                stds = np.zeros(nst)
                for i,tr in enumerate(st_all):
                    stds[i] = tr.corstats[ics].picks[ip].p_final[1]
                median, median_std = _median_std(stds)
                q12 = np.less(stds, 0.3).sum() / float(nst)
                sss += ' #{:d} {:.2f} {:3.0f}%'.format(ip+1, median, 100*q12)
                if q12 >= ibest[2] + edge:
                    ibest = [ics, ip, q12, median]
                elif (abs(q12-ibest[2])<=edge) and (median < ibest[3]):
                    ibest = [ics, ip, q12, median]
            print(sss)
        if ibest[0] < 0:
            print('-> No good peaks')
            if iphase == 1:
                output = os.path.join(fold_out, 'pick_'+id_origin+'_'+phase+'.txt')
                open(output, 'w').close()
            iphase -= 1
            continue
        else:
            print('-> {:d} peaks, {:s} pick'.format(ibest[0]+2, ordinal(ibest[1]+1)))
            ics = ibest[0]
            max_maxmin = len(medstats[ics].picks)
            theo_dif = medstats[ics].wave_start - medstats[ics].trace.stats.p_theor
            wave_start = medstats[ics].wave_start + wave_start_shifts[ics]
            if (nphase == 1) and (theo_dif > max_theo_dif):
                nphase = 2
                phase1 = phase + 'X'
            elif (iphase == 2) and (theo_dif <= max_theo_dif):
                phase1 = phase
                iphase = 0
            else:
                iphase -= 1

            if io_figure:
                tl1 = np.around(medstats[ics].wave_start - 5 - 1)
                tl1 = min([10, tl1])
                tl1 = max([0, tl1])
                tl2 = tl1 + 15

                fig = plt.figure()
                ax = fig.add_subplot(111)
                ii = np.where((tm >= medstats[ics].wave_start) & (tm <= medstats[ics].roots[max_maxmin-1]))[0]
                for tr in st_all:
                    tr2 = tr.corstats[ics].trace
                    tm2 = tr2.times()
                    tr2_signal = tr2.data / np.linalg.norm(tr2.data[ii])
                    ax.plot(tm2, tr2_signal, label=tr2.id)
                ax.plot(tm2, medstats[ics].trace.data, 'k-')
                std_old = np.log10(np.absolute(np.divide(medstats[ics].trace.std, medstats[ics].trace.data)))
                ylims_old = [np.floor(np.nanmin(std_old)), np.ceil(np.nanmax(std_old))]
                ylims_new = ax.get_ylim()
                std_new = (ylims_new[1]-ylims_new[0])/(ylims_old[1]-ylims_old[0]) * \
                        (std_old-ylims_old[0]) + ylims_new[0]
                ax.plot(tm2, std_new, 'k:')
                ax.axvline(x=medstats[ics].wave_start, color='r')
                ax.axvline(x=medstats[ics].roots[max_maxmin-1], color='r')
                ax.set_xlim((tl1,tl2))
                #ax.legend(loc='upper left')
                output_fig = os.path.join(fold_out_pics_evt, id_origin+'_shortcor_all_'+phase1+'.png')
                fig.savefig(output_fig)
                plt.close(fig)

                fig = plt.figure()
                ax = fig.add_subplot(111)
                tm = medstats[ics].trace.times()
                ax.plot(tm, medstats[ics].trace.data, 'k-')
                ii = np.where((tm >= medstats[ics].wave_start) & (tm <= medstats[ics].roots[max_maxmin-1]))
                ax.plot(tm[ii], medstats[ics].trace.data[ii], 'b-')
                #ax.axvline(x=start_trig, color='r')
                #ax.axvline(x=medstats[ics].wave_start, color='b')
                for root in medstats[ics].roots:
                    ax.plot(root, 0, 'ro')
                for maxmin in medstats[ics].maxmins:
                    i = np.where(tm >= maxmin)[0][0]
                    ax.plot([maxmin, maxmin], [0, medstats[ics].trace.data[i]], 'g-')
                ii = np.where((tm >= medstats[ics].noise_level[2]) & (tm <= medstats[ics].noise_level[3]))
                tt1 = tm[ii][0]; tt2 = tm[ii][-1]
                ax.plot([tt1,tt2], [medstats[ics].noise_level[1], medstats[ics].noise_level[1]], 'r-')
                ax.plot([tt1,tt2], [medstats[ics].noise_level[1]-medstats[ics].noise_level[0], 
                            medstats[ics].noise_level[1]-medstats[ics].noise_level[0]], 'r--')
                ax.plot([tt1,tt2], [medstats[ics].noise_level[1]+medstats[ics].noise_level[0], 
                            medstats[ics].noise_level[1]+medstats[ics].noise_level[0]], 'r--')
                _draw_pick(ax, medstats[ics].wave_start, text=phase1+'_abs', color='black', zorder=4)
                # figure xlims
                ax.set_xlim((tl1, tl2))
                # theoretical phases
                x0 = [arrival.time for arrival in arrivals if arrival.name == phase][0]
                for arrival in arrivals:
                    x = medstats[ics].trace.stats.p_theor + arrival.time - x0
                    if x <= tl2:
                        _draw_pick(ax, x, text=arrival.name, color='orange', zorder=4)
                # save figure
                output_fig = os.path.join(fold_out_pics_evt, id_origin+'_shortcor_peaks_'+phase1+'.png')
                fig.savefig(output_fig)
                plt.close(fig)


            #--- OUTPUT ------------------
            print('')
            output = os.path.join(fold_out, 'pick_'+id_origin+'_'+phase1+'.txt')
            f = open(output, 'w')
            for tr in st_all:
                for i in range(max_maxmin):
                    dp_final = (tr.corstats[ics].trace.stats.starttime - tr.stats.p_theor) + tr.corstats[ics].picks[i].p_final[0]
                    dp_final -= (medstats[ics].picks[i].p_maxim[0] - medstats[ics].wave_start)
                    p_theor, io = _round_time(time=tr.stats.p_theor, sample_rate=sample_rate)
                    sss = '{:3d} {:12s}  {:<16s} {:6s} {:s} {:4.0f}%  {:2d} {:5.1f} {:7.2f} {:5.2f}'.format(
                        i_evt, origin.time.strftime("%y%m%d%H%M%S"), tr.id, phase1, p_theor.isoformat()[:22],
                        100*tr.corstats[peak_max1-peak_min].cor_max[-1], i+1, 
                        tr.corstats[ics].picks[i].p_maxim[2], dp_final, tr.corstats[ics].picks[i].p_final[1])
                    print(sss)
                    f.write(sss+'\n')
            f.close()


            #--- OUTPUT PICK FIGS ------------------
            tl1 = np.around(medstats[ics].wave_start - 5 - 1)
            tl1 = min([10, tl1])
            tl1 = max([0, tl1])
            tl2 = tl1 + 15

            if io_figure:
                print('')
                for tr in st_all:
                    fig = plt.figure()
                    ax = fig.add_subplot(111)
                    tm = tr.corstats[ics].trace.times()
                    ii = np.where((tm >= medstats[ics].wave_start) & (tm <= medstats[ics].roots[max_maxmin-1]))[0]
                    for tr_all in st_all:
                        tr2 = tr_all.corstats[ics].trace
                        tm2 = tr2.times()
                        tr2_signal = tr2.data / np.linalg.norm(tr2.data[ii])
                        ax.plot(tm2, tr2_signal, c='silver', zorder=1)
                    ax.plot(tm, medstats[ics].trace.data, c='black', zorder=2)
                    tr1 = tr.corstats[ics].trace
                    norm1 = np.linalg.norm(tr1.data[ii])
                    tr1_signal = tr1.data / norm1
                    ax.plot(tm, tr1_signal, c='red', zorder=3)
                    # noise level
                    color = 'm'
                    tt1 = tr.corstats[ics].noise_level[2]; tt2 = tr.corstats[ics].noise_level[3]
                    median = tr.corstats[ics].noise_level[1] / norm1; std_med = tr.corstats[ics].noise_level[0] / norm1
                    ax.plot([tt1,tt2], [median-std_med, median-std_med], c = color, ls = '--', zorder=3)
                    ax.plot([tt1,tt2], [median+std_med, median+std_med], c = color, ls = '--', zorder=3)
                    # top cup
                    color = 'cyan'
                    for i in range(max_maxmin):
                        ii = tr.corstats[ics].picks[i].p_top_range
                        yn = 0.5 * (tr1_signal[ii[0]] + tr1_signal[ii[-1]])
                        ax.fill_between(tm[ii], yn, tr1_signal[ii], facecolor=color, edgecolor=color, interpolate=True, zorder=8)
                    # set axes limits
                    ax.set_xlim((tl1, tl2))
                    ii = np.where((tm >= tl1) & (tm <= tl2))[0]
                    y1, y2 = ax.get_ylim()
                    ax.set_ylim((max([y1, 2*np.amin(medstats[ics].trace.data[ii])]), 
                         min([y2, 2*np.amax(medstats[ics].trace.data[ii])])))
                    # draw picks
                    _draw_pick(ax, medstats[ics].wave_start, text=phase1+'_abs', color='black', zorder=4)
                    for i in range(max_maxmin):
                        text = '{:s}{:d}'.format(phase1, i+1)
                        _draw_pick(ax, medstats[ics].picks[i].p_maxim[0], color='black', zorder=4)
                        _draw_pick(ax, tr.corstats[ics].picks[i].p_correl[0], color='blue', zorder=5)
                        _draw_pick(ax, tr.corstats[ics].picks[i].p_maxim[0], color='red', zorder=6)
                        _draw_pick(ax, tr.corstats[ics].picks[i].p_final[0], text=text, color='green', zorder=7)
                    ax.set_xlabel('Time after {:s} [s]'.format(tr.corstats[ics].trace.stats.starttime.isoformat()[:22]))
                    ax.set_title(tr.id)
                    plt.show()

                    for ip in range(max_maxmin):
                        print('#{:d} {:15s} red: {:.2f} {:.2f} > {:.2f}  black: {:.2f} {:3.0f}% {:.2f} > {:.2f}  blue: {:.2f} {:3.0f}% {:.2f} > {:.2f}  green: {:.2f} {:.2f}'.format(
                            ip+1, tr.id, tr.corstats[ics].picks[ip].p_maxim[0], tr.corstats[ics].picks[ip].p_top[2],
                            tr.corstats[ics].picks[ip].p_top[2] + (1-tr.corstats[ics].cor_max[-1])*(ip+1),
                            medstats[ics].picks[ip].p_maxim[0], 100*tr.corstats[ics].cor_max[-1], medstats[ics].picks[ip].p_top[2],
                            medstats[ics].picks[ip].p_top[2] + (1-tr.corstats[ics].cor_max[-1])*(ip+1),
                            tr.corstats[ics].picks[ip].p_correl[0], 100*tr.corstats[ics].picks[ip].p_correl[1], 
                            tr.corstats[ics].picks[ip].p_correl[2], tr.corstats[ics].picks[ip].p_correl[2] + 
                                (1-tr.corstats[ics].picks[ip].p_correl[1]) + (1-tr.corstats[ics].cor_max[-1])*ip,
                            tr.corstats[ics].picks[ip].p_final[0], tr.corstats[ics].picks[ip].p_final[1]))

                    output_fig = os.path.join(fold_out_pics_evt, id_origin+'_'+tr.id+'_picks_'+phase1+'.png')
                    fig.savefig(output_fig)
                    plt.close(fig)



print('HOTOVO')
