#!/usr/bin/env python

# 2017 Ludek Vecsey, IG CAS Prague

# REMARKS:
# 'silo' is a local pc in IG CAS Prague, it is not possible to connect it from elsewhere (I hope)
# to use ALPARRAY and EASI secure, you need to edit _user

"""
USAGE: mseed_router [OPTIONS]

OPTIONS:
 -n --network	= NETWORK	Network code (needed)
 -s --station	= STATION	Station code (needed)
 -l --location	= LOCATION	Location code [default: *]
 -t --channnel	= CHANNEL	Short channel code [default: *]; can be BB or SP for broadband or shortperiod
 -c --component	= COMPONENT	Component code [default: *]
 -d --starttime	= TIME		Start time to read (needed)
 -e --endtime	= TIME		End time to read (needed)
 -o --outdir	= FOLDER	Folder to save mseeds [default: .]
 -r --routefile	= FILE		File with list of routes to mseed archives [default: data_archives.txt]
 -g --logfile	= FILE		File to append logs (if not given, no log output to file) [default: None]
 -v --verbose	= 0|1|2|3	Verbose level (0:quiet, 3:all)) [default: 0]; only for screen, logfile is not affected
 --firstroute			Do not try other possible routes to mseeds, only first one
 -h --help
 --version
"""

import re, os, sys, glob
import numpy as np
import fnmatch
import math
import datetime
from time import sleep
from optparse import OptionParser
from obspy.core import Stream, UTCDateTime, read
from obspy.arclink import Client

class Stat_routes(object):
    """
    Read and compile file containing routes to data repositories for each station
    """
    
    def __init__(self, routefile='pom.txt'):
        if not os.path.isfile(routefile):
            print('! routes: file {} does not exist'.format(routefile))
            sys.exit(1)
        self.stats = []
        self.routes = []
        with open(routefile) as fr:
            for line in fr:
                # comments
                _i = line.find('#')
                if _i >=0:
                    line = line[:_i]
                # find routes
                line = line.split()
                if len(line) >= 2:
                    self.stats.append(re.compile(fnmatch.translate(line[0])))
                    self.routes.append(line[1])
        # find root silo directory
        self.silo = filter(os.path.isdir, ['/mnt/silo/arch', '/arch'])
        if self.silo: self.silo = self.silo[0]
                    
    def get_route(self, stat):
        route = []
        for i, regex in enumerate(self.stats):
            if regex.match(stat):
                route.append(self.routes[i].lower())
        return route

def stations_in_network(network):
    """
    List all stations from given network
    examples: 'XT' all the network, 'XT:CZ' only Czech stations
    """
    stations = []
    network = network.split(':')
    if network[0] == 'CZ':
        stations = ['JAVC', 'KRUC', 'MORC', 'VRAC', 'DPC', 'GOPC', 'KHC', 'KRLC', 'NKC', 'PBCC', 'PRU', 
                    'PVCC', 'TREC', 'CHVC', 'UPC', 'OSTC', 'CKRC', 'PRA', 'OKC', 'HSKC']
    elif network[0] == 'XT':
        n1 = 1; n2 = 55
        try:
            if network[1] == 'CZ': 
                n2 = 20
        except Exception:
            pass
        [stations.append("AAE{0:02d}".format(i)) for i in range(n1,n2+1)]
    elif network == 'Z3':
        # not complete yet
        n1 = 1; n2 = 369
        try:
            if network[1] == 'CZ': 
                n1 = 79
                n2 = 90
        except Exception:
            pass
        [stations.append("A{0:03d}A".format(i)) for i in range(n1,n2+1)]
    
    return stations

def human_time(secs):
    units = [("d", 86400), ("h", 3600), ("m", 60), ("s", 1)]
    parts = ""
    for unit, mul in units:
        if secs / mul >= 1 or mul == 1:
            if mul > 1:
                n = int(math.floor(secs / mul))
                secs -= n * mul
            else:
                n = secs if secs != int(secs) else int(secs)
            part = "{0}{1}".format(n, unit)
            if part == "0s": part = ""
            parts += part
    return parts

def _access_logfile(logfile, status=None):
    flog = None
    if logfile:
        _fold = os.path.dirname(logfile)
        if not os.path.exists(_fold):
            try:
                os.makedirs(_fold)
            except Exception:
                pass
        flog = open(logfile, 'a+')
        if status == 'init':
            _now = str(datetime.datetime.now())
            if os.path.getsize(logfile) > 0:
                flog.write('\n')
            flog.write('### {0} ###\n'.format(_now))
            flog.close()
    return flog

def _get_times(starttime, endtime, sformat):
    _t1 = UTCDateTime(starttime.strftime('%Y-%m-%d'))
    _t2 = UTCDateTime(endtime.strftime('%Y-%m-%d'))
    _days = int((_t2 - _t1) / 86400.)
    _tt = [_t1 + 86400. * x for x in range(0, _days+1)]
    if sformat == 'yymmdd':
        stimes = [x.strftime('%y%m%d') for x in _tt]
    elif sformat == 'yyyy/jul':
        stimes = [os.path.join(x.strftime('%Y'), x.strftime('%j')) for x in _tt]
    return stimes

def _disp_mesg(text_or_stream, flog, verbose, flag=0):
    if flag == 0:
        try:
            text_or_stream.sort()
            _txt = "   " + text_or_stream.__str__(extended=True).split('\n',1)[-1].replace('\n','\n   ')
        except Exception:
            _txt = text_or_stream
        if flog:
            flog.write(_txt+'\n')
        if verbose > 0:
            print(_txt)
    return

def _st_completness(_st, _delta0):
    _ids = []
    _rdelta = []
    for _tr in _st:
        _ids.append(_tr.id)
    _ids = sorted(set(_ids))
    for _id in _ids:
        _st1 = _st.select(id=_id)
        _delta = 0.
        for _tr1 in _st1:
            _delta += _tr1.stats.endtime - _tr1.stats.starttime
        _rdelta.append(100. * _delta / _delta0)
    return _ids, _rdelta

def _heal_by_extrfile(filename, sourcename, key=None):
    
    extr_file = 'extr_file'
    pom = 'pom.txt'
    _st = Stream()
    if key == 'ascii':
        ef_add = '-f c'  # forced STEIM1
    else:
        ef_add = ''
    ef_add += ' -n o'    # not to have troubles with space network in filenames
    try:
        os.system("{0} {1} {2:s} > {3}".format(extr_file, filename, ef_add, pom))
        with open(pom) as f:
            for line in f.readlines():
                if 'written on file' in line:
                    _mseed = line.rsplit(' ', 1)[1].rstrip()
                    break
        try:
            _st = read(_mseed, format='MSEED', sourcename=sourcename)
        except Exception:
            pass
        os.remove(pom)
        os.remove(_mseed)
    except Exception:
        pass
    return _st

def _get_mseed_silo(filename, starttime, endtime, netstatchan, flog, verbose):
    """
    1. if mseed not readable, try extr_file first, then read again
    2. if ascii mseed (encoding=0), probably only bad info about encoding in blockette 1000,
       heal it by extr_file
    """
    
    _st = Stream()
    _st_fin = Stream()
    _netstatchan = '*.' + netstatchan.split('.', 1)[1]
    
    try:
        _st = read(filename, format='MSEED', sourcename=_netstatchan)
        for _tr in _st:
            if _tr.stats.mseed['encoding'] == 'ASCII':
                _st = _heal_by_extrfile(filename, sourcename=_netstatchan, key='ascii')
                if _st:
                    _disp_mesg('?  {0} | silo | ascii {1} healed by extr_file'.format(
                        _netstatchan, os.path.basename(filename)), flog, verbose)
                break
    except Exception as e:
        _st = _heal_by_extrfile(filename, sourcename=_netstatchan)
        if _st:
            _disp_mesg('?  {0} | silo | {1} healed by extr_file'.format(
                    _netstatchan, os.path.basename(filename)), flog, verbose)
    
    if not _st:
        if verbose > 2:
            print(e)
            if flog: flog.write(str(e)+'\n')
    else:
        _st_fin = _st.select(network=netstatchan.split('.', 1)[0])
        if not _st_fin:
            _netstatchan = _st[0].stats.network +'.'+_st[0].stats.station+'.'+ \
                               _st[0].stats.location+'.'+_st[0].stats.channel
            _tspan = human_time(endtime-starttime)
            _disp_mesg('?  {0} | silo | {1} {2} | found {3} instead'.format(
                    netstatchan, starttime, _tspan, _netstatchan), flog, verbose)
            _st_fin = _st
        _st_fin = _st_fin.slice(starttime=starttime, endtime=endtime)
    return _st_fin

def _get_mseed_eida(host, port, user, starttime, endtime, netstatchan, flog, verbose):
    """
    Try arclink several times (see n_repeat) if connection failed
    """
    if not user:
        user = "gfu@ig.cas.cz"   # feel free to change it
    n_repeat = 3
    _st = Stream()
    _netstatchan = netstatchan.split('.')
    _tspan = human_time(endtime-starttime)
    
    client = Client(host=host, port=port, user=user)
    for _i in range(n_repeat):
        try:
            _st = client.getWaveform(network=_netstatchan[0], station=_netstatchan[1],
                                    location=_netstatchan[2], channel=_netstatchan[3],
                                    starttime=starttime, endtime=endtime)
            if len(_st) == 0: se = 'No data available'
        except Exception as e:
            se = str(e)
        if _st or (se == 'No data available') or ('Could not find route' in se):
            se = ''
            break
        if 'denied' in se.lower():
            break
        _disp_mesg('{0:d}  {1} | eida | {2} {3} | {4}'.format(_i+1, netstatchan, starttime, _tspan, se), 
                   flog, verbose-1)
        sleep(2)
    if se:
        _disp_mesg('?  {0} | eida | {1} {2} | {3}'.format(netstatchan, starttime, _tspan, se), 
                   flog, verbose)
    client._client.close()
    return _st
    
def _corrections(st, flog=None, verbose=0):
    _mesg = []
    for tr in st:
        # no network
        if (not tr.stats.network) and (tr.stats.station in ['PRU']):
            _id_old = tr.id
            tr.stats.network = 'CZ'
            _id_new = tr.id
            _mesg.append('?  {0} | changed to {1}'.format(_id_old, _id_new))
        # small deviation in sampling rate number
        sr = np.round(tr.stats.sampling_rate)
        sratio = np.absolute(tr.stats.sampling_rate - sr) / sr
        if (sratio < 1e-6) and (tr.stats.sampling_rate != sr):
            _mesg.append('?  {0} | {1:.6f} Hz changed to {2:.1f} Hz'.format(tr.id, tr.stats.sampling_rate, sr))
            tr.stats.sampling_rate = sr
    _mesg = sorted(set(_mesg))
    for _m in _mesg:
        _disp_mesg(_m, flog, verbose)
    return st

def _get_signal(route, net, stat, loc, chan, comp, starttime, endtime, flog, verbose=0):
    
    global statroutes
    keys = ['perm', 'mobnet']
    
    if np.mod(verbose,1) > 0:
        io_verbose = 1
        verbose = int(verbose)
    else:
        io_verbose = 0
    
    _st = Stream()
    netstatchan = net+'.'+stat+'.'+loc+'.'+chan+comp
    netstatchan = netstatchan.replace('**', '*')
    _route = route.split(':')
    _event = starttime.isoformat()
    _delta0 = endtime - starttime
    _tspan = human_time(_delta0)
    _ident = "{0} | {1} | {2} {3}".format(netstatchan, route, _event, _tspan)
    
    _disp_mesg(_ident+" | looking for data", flog, verbose-1, flag=io_verbose)
    
    if _route[0] == 'silo':     # GFU silo
        if len(_route) == 1:
            for key in keys:
                _st += _get_signal(route=_route[0]+':'+key, net=net, stat=stat, loc=loc, chan=chan, 
                            comp=comp, starttime=starttime, endtime=endtime, flog=flog, verbose=verbose)
        else:
            
            if _route[1] == 'mobnet':     # MOBNET
                _fold = os.path.join(statroutes.silo, 'mobnet', 'MOBNET_MSEED', stat, 'data_mseed')
                # ALPARRAY temporary directory
                if net == 'Z3':
                    _fold = os.path.join('/mnt/silo/bohema/ALPARRAY/MSEED', stat)
                # ---
                sdates = _get_times(starttime, endtime, 'yymmdd')
                _files = []
                for sdate in sdates:
                    _files.extend(glob.glob(os.path.join(_fold, stat+sdate+'*.'+chan+comp)))
#                    _st += _get_mseed_silo(_file, starttime, endtime, netstatchan, flog, verbose)
            elif _route[1] == 'perm':     # PERMANENT
                _fold = os.path.join(statroutes.silo, 'service', 'gfu')
                sdates = _get_times(starttime, endtime, 'yyyy/jul')
                _files = []
                for sdate in sdates:
                    _files.extend(glob.glob(os.path.join(_fold, sdate, stat+'.'+chan+comp+'.*')))
            # get data from silo
            _files = sorted(set(_files))
            for _file in _files:
                _st += _get_mseed_silo(_file, starttime, endtime, netstatchan, flog, verbose)
                
    elif _route[0] == 'eida':   # EIDA via arclink
        _user = None
        if len(_route) == 1:           _host = 'webdc.eu'; _port = '18002'              # EIDA
        else:
            if   _route[1] == 'odc':   _host = 'eida.knmi.nl'; _port = '18002'          # ODC Orfeus
            elif _route[1] == 'gfz':   _host = 'eida.gfz-potsdam.de'; _port = '18001'   # GFZ
            elif _route[1] == 'resif': _host = 'eida.resif.fr'; _port = '18001'         # RESIF
            elif _route[1] == 'ingv':  _host = 'eida.ingv.it'; _port = '18002'          # INGV
            elif _route[1] == 'ethz':  _host = 'eida.ethz.ch'; _port = '18001'          # ETHZ
            elif _route[1] == 'bgr':   _host = 'eida.bgr.de'; _port = '18001'           # BGR
            elif _route[1] == 'lmu':                                                    # LMU
                _host = 'erde.geophysik.uni-muenchen.de'; _port = '18001'
            elif _route[1] == 'easi':                                                   # EASI secure
                _host = 'arclink.ethz.ch'; _port = '18002'; _user = ''
            elif _route[1] == 'alparray':                                               # ALPARRAY secure
                #_host = 'webdc.eu'; _port = '18002'; 
                _user = ''
                _host = 'eida.knmi.nl'; _port = '18002'
            else:                      _host = 'webdc.eu'; _port = '18002'              # EIDA
        _st += _get_mseed_eida(_host, _port, _user, starttime, endtime, netstatchan, flog, verbose)
        
    
    # FINAL MESSAGE

    if _st:
        _st.merge(method=-1)
        _st = _st.split()
        _disp_mesg(_st, flog, verbose-1)
        _ids, _rdelta = _st_completness(_st, _delta0)
        for i, _id in enumerate(_ids):
            if _rdelta[i] < 80.:
                _disp_mesg('?  {0:s} | {1} | {2} {3} | only {4:.0f} % of data'.format(
                        _id, route, _event, _tspan, _rdelta[i]), flog, verbose)
            else:
                _disp_mesg('-> {0:s} | {1} | {2} {3} | got {4:.0f} % of data'.format(
                        _id, route, _event, _tspan, _rdelta[i]), flog, verbose)
    else:
        _disp_mesg('!  {} | not found'.format(_ident), flog, verbose-0, flag=io_verbose)
        
    return _st

def read_mseed_route(network, station, location='*', channel='*', component='*', 
                    starttime=UTCDateTime()-3600, endtime=UTCDateTime(),
                    outdir=None, logfile=None, routefile='data_archives.txt',
                    verbose=0, check_all_archives=True):
    """
    network  ... probably can be with wild cards but it was not tested; list of networks: no
    station  ... no wild cards or list!
    location ... wild cards: yes, list of locations: yes. Examples: '*', ['01','05']
    channel  ... wild cards: yes, list: yes. Examples: 'LH', ['BH','SH'], '?H', '*'
                 specials: 'BB' means 'HH' (or 'BH' if HH does not exist)
                           'SP' means 'EH' (or 'SH' if EH does not exist)
    component ... wild cards: yes, list: yes. Examples: 'Z', ['N','E'], '*'
                  If no of '1','2','3' is included in component, it automatically tests for them
    """
    
    global statroutes

    # read all routes (if not already done)
    try:
        if not statroutes: statroutes = Stat_routes(routefile)
    except Exception:
        statroutes = Stat_routes(routefile)
    
    # inicialization
    st = Stream()
    
    network = network.upper()
    station = station.upper()
    if isinstance(location, basestring): location = [location]
    if isinstance(channel, basestring): channel = [channel]
    channel = [x.upper() for x in channel]
    component = [x.upper() for x in component]
    netstat = '{0}.{1}'.format(network, station)
    
    comp_123 = ['1', '2', '3']
    io_comp_123 = [i for i in component if i in comp_123]
    io_component = False
    io_component_verb = False
    if (not io_comp_123) and (not '*' in component):
        if len(component) == 1:
            io_component = True
        io_component_verb = True
        component.extend(comp_123)
    if '*' in component:
        component = ['*']
        
    
    # access log file
    flog = _access_logfile(logfile)
    
    # find routes to station data archives
    routes = statroutes.get_route(netstat)
    if not routes:
        _disp_mesg('!  {0} | no known route to archive'.format(netstat), flog, verbose)
        if flog: flog.close()
        return st
    
    if not check_all_archives:
        routes = routes[0:1]
    
    # loops
    for loc in location:
        _st_comp = Stream()
        for comp in component:
            if io_component_verb and (comp in comp_123):
                _verbose = verbose + 0.5
            else:
                _verbose = verbose
            for chan in channel:
                if chan == 'SP':
                    _chans = ['EH', 'SH']
                elif chan == 'BB':
                    _chans = ['HH', 'BH']
                else:
                    _chans = [chan]
                for _chan in _chans:
                    for route in routes:
                        _st_chan = _get_signal(route=route, net=network, stat=station, loc=loc, chan=_chan, comp=comp,
                                      starttime=starttime, endtime=endtime, flog=flog, verbose=_verbose)
                        if _st_chan: 
                            _st_comp += _st_chan
                            break
                    if _st_chan: break
            if _st_comp and io_component: break
        st += _st_comp
        
    if st:
        st.merge(method=-1)
        st = st.split()
        # correct known issues
        st = _corrections(st, flog=flog, verbose=_verbose)
        # output
        if outdir:
            if not os.path.isdir(outdir):
                os.makedirs(outdir)
            _ids = []
            for tr in st:
                _ids.append(tr.id)
                if tr.data.dtype != 'int32':
                    tr.data = np.around(tr.data)
                    tr.data = tr.data.astype("int32")
            _ids = sorted(set(_ids))
            _event = starttime.strftime('%y%m%d%H%M%S')
            for _id in _ids:
                st1 = st.select(id=_id)
                if not _id.split('.', 1)[0]:
                    __id = '__.' + _id.split('.', 1)[1]
                else:
                    __id = _id
                _file = os.path.join(outdir, __id + '.' + _event + '.mseed')
                st1.write(_file, format='MSEED', encoding='STEIM1')
    
    # finish
    if flog: flog.close()
    return st

def main(option_list=None):

    global statroutes

#    parser = OptionParser(__doc__.strip())
    parser = OptionParser(usage="usage: %prog [options]", version="%prog 1.0")

    parser.add_option("-n", "--network", action="store", 
                      help="network code; example: CZ")
    parser.add_option("-s", "--station", action="store", 
                      help="station code; example: PRU")
    parser.add_option("-l", "--location", action="store", default="*",
                      help="location code [default: *]; examples: 00, '01 05', '0[1-3]', '1?'")
    parser.add_option("-t", "--channel", action="store", default="*",
                      help="short channel code [default: *]; "
                           "BB or SP for broadband (HH/BH) or shortperiod (EH/SH); "
                           "examples: HH, '?H', 'SP BB LH'")
    parser.add_option("-c", "--component", action="store", default="*",
                      help="component [default: *]; if not given, automatically check for '1 2 3'; "
                      "examples: Z, 'N E', 'Z 1'")
    parser.add_option("-d", "--starttime", action="store",
                      help="start time to read (in format as for UTCDateTime); "
                      "examples: 2009-12-31T12:23:34.5, 2009-365T12:23:34.5, '1970-01-01 12:23:34'")
    parser.add_option("-e", "--endtime", action="store",
                      help="end time to read (in format as for UTCDateTime); "
                      "examples: 2009-12-31T12:23:34.5, 20090701121212, 2009365T122334.5")
    parser.add_option("-o", "--outdir", action="store", default=".",
                      help="folder to save mseeds [default: .]; example: ./MSEEDS")
    parser.add_option("-r", "--routefile", action="store", default="data_archives.txt",
                      help="file with list of routes to mseed archives [default: data_archives.txt]")
    parser.add_option("-g", "--logfile", action="store",
                      help="file with appended logs (if not given, no logs go to file) [default: None]")
    parser.add_option("-v", "--verbose", action="store", type="int",
                      help="verbose level (0:quiet, 3:all reports) [default: 0]; "
                      "affects logs displayed on screen only, not in logfile")
    parser.add_option("--firstroute", action="store_false", default=True, dest="check_all_archives",
                      help="do not follow other possible routes to mseeds, only first one")
    (options, largs) = parser.parse_args(option_list)

#    print(options)
#    print(largs)

    # Print help and exit if no arguments are given
    if not options.network or not options.station or \
       not options.starttime or not options.endtime:
        parser.print_help()
        sys.exit(1)

    # Check inputs
    try:
        verbose = options.verbose
        if verbose < 0: verbose = 0
        starttime = UTCDateTime(options.starttime)
        endtime = UTCDateTime(options.endtime)
        loc = options.location.split()
        chan = options.channel.split()
        comp = options.component.split()
    except Exception as e:
        print(e)
        sys.exit(1)

    # Access log file
    _access_logfile(options.logfile, status='init')

    # Read mseed
    statroutes = None
    st = read_mseed_route(options.network, options.station, location=loc, channel=chan, 
                          component=comp, starttime=starttime, endtime=endtime, 
                          outdir=options.outdir, logfile=options.logfile, verbose=verbose,
                          routefile=options.routefile, check_all_archives=options.check_all_archives)


if __name__ == '__main__':
    main()
