#!/usr/bin/env python

"""
NEW UTILS:

dist_baz	... epicentral distance, azimuth and backazimuth (all in degrees)
arrival_incid	... arrival time(s) and incidence angle(s)
better_demean	... gaps will not cause separate means in one trace
better_merge	... small gaps are interpolated, large gaps filled by zero
simul_wwssn_sp	... wwssn sp simulation

Ludek Vecsey, IG CAS Prague, 28.3.2017
"""

import os
from obspy.station import read_inventory, Inventory
from obspy.core import Stream, UTCDateTime, read
from obspy.core.util.geodetics.base import calcVincentyInverse, kilometer2degrees
from obspy.taup import TauPyModel
from obspy.xseed import Parser
import numpy as np

def dist_baz(stat_lat, stat_lon, evt_lat, evt_lon):
    """
    Computes epicentral distance, azimuth and backazimuth (all in degrees)
    """
    [dist_m, az, baz] = calcVincentyInverse(evt_lat, evt_lon, stat_lat, stat_lon)
    dist = kilometer2degrees(dist_m/1000.)
    return dist, az, baz

def arrival_incid(source_depth, distance, phases="P-first", model=u'iasp91'):
    """
    Computes arrival time(s) and incidence angle(s)
    
    INPUTS:
       source_depth ... in meters
       distance     ... in degrees
       phases examples: 'P', 'SKSac', 'P-first' = first primary, ['pS', 'Pdiff']
       model        ... 'iasp91' [default], 'ak135'
    OUTPUTS:
       phase_list   ... names of phases
       offset_list  ... time offsets from event.origin
       incid_list   ... list of incidence angles (in degrees)
    
    Available phases:
                   P, P'P'ab, P'P'bc, P'P'df, PKKPab, PKKPbc, PKKPdf, PKKSab,
                   PKKSbc, PKKSdf, PKPab, PKPbc, PKPdf, PKPdiff, PKSab, PKSbc,
                   PKSdf, PKiKP, PP, PS, PcP, PcS, Pdiff, Pn, PnPn, PnS, S,
                   S'S'ac, S'S'df, SKKPab, SKKPbc, SKKPdf, SKKSac, SKKSdf,
                   SKPab, SKPbc, SKPdf, SKSac, SKSdf, SKiKP, SP, SPg, SPn, SS,
                   ScP, ScS, Sdiff, Sn, SnSn, pP, pPKPab, pPKPbc, pPKPdf,
                   pPKPdiff, pPKiKP, pPdiff, pPn, pS, pSKSac, pSKSdf, pSdiff,
                   sP, sPKPab, sPKPbc, sPKPdf, sPKPdiff, sPKiKP, sPb, sPdiff,
                   sPg, sPn, sS, sSKSac, sSKSdf, sSdiff, sSn
    """
    _model = TauPyModel(model=model)
    try:
        if phases == 'P-first':
            arrivals = _model.get_travel_times(source_depth_in_km=source_depth/1000., 
                                               distance_in_degree=distance)
            arrivals = [arrivals[0]]
        else:
            if isinstance(phases, basestring): phases = [phases]
            arrivals = _model.get_travel_times(source_depth_in_km=source_depth/1000., 
                                               distance_in_degree=distance, phase_list=phases)
    except Exception as e:
        print(e)
        arrivals = []
        
    phase_list, offset_list, incid_list = [], [], []
    for i, arrival in enumerate(arrivals):
        phase_list.append(arrival.name)
        offset_list.append(arrival.time)
        incid_list.append(arrival.incident_angle)
    return phase_list, offset_list, incid_list

def better_demean(st, method='mean'):
    """
    Demean all traces in a stream, gaps will not cause separate means in one trace
    INPUT:
       st     ... stream
       method ... 'mean', 'median'
    OUTPUT:
       st_demean
    """
    st_demean = st.copy()
    ids = []
    for tr in st_demean:
        ids.append(tr.id)
    ids = sorted(set(ids))
    for _id in ids:
        _st = st_demean.select(id=_id)
        data = []
        for _tr in _st:
            data.extend(_tr.data)
        data = np.array(data)
        if method == 'mean':
            koef = np.mean(data)
        elif method == 'median':
            koef = np.median(data)
        else:
            disp('{0} not implemented!'.format(method))
            koef = 0.
        for _tr in _st:
            _tr.data -= int(round(koef))
    return st_demean

def better_merge(st, edge_gap=5, demedian=True):
    """
    Merge traces with same ID together, small gaps are interpolated, large gaps filled by zero
    INPUTS:
        st              ... stream to be merged
        edge_gap=5      ... in sec, see gaps
        demedian = True ... remove median first
    OUTPUT:
        st_nogap        ... stream with no gaps and overlaps
    HOW TO DEAL WITH:    
        overlaps ... same as Stream.merge(method=1)
        gaps     ... if <= edge_gaps, then fill_value = 'interpolation'
                                      else fill_value = 0
    """
    if demedian:
        st = better_demean(st, 'median')
    st_nogap = Stream()
    
    ids = []
    for tr in st:
        ids.append(tr.id)
    ids = sorted(set(ids))
    
    for _id in ids:
        st_id = st.select(id=_id)
        st_id_nogap = st_id.copy().sort().merge(method=1).split()
        gaps = st_id_nogap.getGaps()
        if gaps:
            for gap in gaps:
                if gap[6] <= edge_gap:
                    fill_method = 'interpolate'
                else:
                    fill_method = 0
                st12 = Stream([st_id_nogap[0], st_id_nogap[1]])
                st12.merge(method=1, fill_value=fill_method).trim(endtime=gap[5])
                st_id_nogap.pop(0)
                st_id_nogap = st_id_nogap.append(st12[0]).merge(method=1).split()
        st_nogap.append(st_id_nogap[0])
        
    return st_nogap

def better_slice(tr, starttime, endtime):
    """
    Slice of trace with accurate number of samples
    """
    npt = int(np.round((endtime - starttime) * tr.stats.sampling_rate)) + 1
    tr_new = tr.slice(starttime, endtime + 2 * tr.stats.delta)
    if tr_new.stats.npts > npt:
        tr_new.data = tr_new.data[:npt]
    return tr_new

def simul_wwssn_sp(st, metadata, demedian=False, output="DISP"):
    """
    Simulation of WWSSN-SP seismometer, preventation of overamplification by Butterworth BB zero-phase filter
    INPUTS:
        st       ... stream
        metadata ... one of:
                     [file].dless ... name of dataless seed file [string]
                     [file].xml   ... name of StationXML file [string]
                     parser       ... dataless seed parser [class: parser]
                     inventory    ... StationXML inventory [class: inventory]
        demedian ... if True, it will remove a median from all traces in the stream first
        output   ... output units, one of:
                     "DISP" ... displacement, output unit is meters
                     "VEL"  ... velocity, output unit is meters/second
                     "ACC"  ... acceleration, output unit is meters/second**2
    OUTPUTS:
        st_simul ... stream
    """
    
    paz_wwssn_sp = {'poles': [-3.367788+3.731514j, -3.367788-3.731514j,
                              -7.037168+4.545562j, -7.037168-4.545562j],
                    'zeros': [0j, 0j],
                    'gain': 70.18386,
                    'sensitivity': 1.}
    
    if demedian:
        st_simul = better_demean(st, 'median')
    else:
        st_simul = st.copy()
        
    #print(st[0].stats)
    #print(st_simul[0].stats)
        
    error = None
    if isinstance(metadata, basestring):
        extension = os.path.splitext(metadata)[1]
        if extension in ['.dless', '.seed']:
            try:
                metadata = Parser(metadata)
            except Exception as e:
                error = str(e)
        elif extension == '.xml':
            try:
                metadata = read_inventory(metadata)
            except Exception as e:
                error = str(e)
        else:
            error = 'Can not recognize format of {}'.format(metadata)
    
    #if isinstance(metadata, obspy.xseed.parser.Parser):
    if isinstance(metadata, Parser):
        try:
            st_simul.simulate(seedresp={'filename': metadata, 'units': output[:3]}, 
                              paz_simulate=paz_wwssn_sp)
        except Exception as e:
            error = str(e)
    #elif isinstance(metadata, obspy.station.inventory.Inventory):
    elif isinstance(metadata, Inventory):
        try:
            st_simul.attach_response(metadata)
            st_simul.remove_response(zero_mean=False, taper=False, water_level=1e6)
            st_simul.simulate(paz_remove=None, paz_simulate=paz_wwssn_sp)
        except Exception as e:
            error = str(e)
            
    if not error:
        # Aditional bandpass filter
        st_simul.filter('bandpass', freqmin=0.2, freqmax=5., corners=2, zerophase=True)
        #pass
    else:
        print(error)
        
    return st_simul
